import asyncio
import json
import math
import multiprocessing
import tempfile
from typing import Tuple, List
import ffmpeg
import os
import uuid
import requests
import sys
import time
import struct
import shutil

from pyrtmp import RTMPProtocol, StreamClosedException
from pyrtmp.messages import SessionManager
from pyrtmp.messages.audio import AudioMessage
from pyrtmp.messages.command import NCConnect, NCCreateStream, NSPublish, NSCloseStream, NSDeleteStream
from pyrtmp.messages.data import MetaDataMessage
from pyrtmp.messages.protocolcontrol import SetPeerBandwidth, WindowAcknowledgementSize, SetChunkSize
from pyrtmp.messages.usercontrol import StreamBegin
from pyrtmp.messages.video import VideoMessage
from pyrtmp.misc.flvdump import FLVFile, FLVMediaType
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from datetime import datetime
import tracemalloc


sys.path.append('../../')
from rtmp_worker.model.task_status_enum import TaskStatusEnum
from rtmp_worker.model.task_status_response import TaskStatusResponse
from rtmp_worker.model.task_submit_response import TaskSubmitResponse
from rtmp_worker.model.db_models import Words, WordsStats

RTMP_URL = 'rtmp://127.0.0.1/'
API_URL = 'https://api.asya.ai/'

API_KEY = 'a13e02b1-7c86-43dc-a958-c51bd7f7da31'
API_TASK_SUBMIT_URL = API_URL + 'task_submit'
API_TASK_TEXT_STATUS_URL = API_URL + 'task_status'
API_TASK_PRIORITY = 9000
API_CALL_INTERVAL_SEC = 20
API_SUBMIT_TASK_FEATURES = [
#    'audio_denoise',
    'audio_diarisation',
    'audio_text'
]

AUDIO_SAMPLE_RATE = 16000
AUDIO_CHANNELS = 2
AUDIO_FORMAT = 'wav'
AUDIO_CODEC = 'pcm_s16le'
AUDIO_LENGTH_SEC = 60
_AUDIO_BITS_PER_SAMPLE = 16

RELATIVE_UNPROCESSED_FOLDER = '../output/unprocessed'
RELATIVE_PROCESSED_FOLDER = '../output/processed'
RELATIVE_WAITING_FOLDER = '../output/waiting'
RELATIVE_WRITING_FOLDER = '../output/writing'
RELATIVE_DATABASE_FILE = '../database.sqlite3'
SCRIPT_FOLDER = os.path.dirname(os.path.realpath(__file__))
UNPROCESSED_FOLDER = os.path.join(SCRIPT_FOLDER, RELATIVE_UNPROCESSED_FOLDER)
PROCESSED_FOLDER = os.path.join(SCRIPT_FOLDER, RELATIVE_PROCESSED_FOLDER)
WAITING_FOLDER = os.path.join(SCRIPT_FOLDER, RELATIVE_WAITING_FOLDER)
WRITING_FOLDER = os.path.join(SCRIPT_FOLDER, RELATIVE_WRITING_FOLDER)
DATABASE_FILE = os.path.join(SCRIPT_FOLDER, RELATIVE_DATABASE_FILE)

is_running = True


def has_wav_header(wav_contents: bytes):
    is_riff_there: bool = wav_contents[:4].decode('utf-8', errors='ignore') == 'RIFF'
    is_wave_there: bool = wav_contents[8:12].decode('utf-8', errors='ignore') == 'WAVE'
    result: bool = is_riff_there and is_wave_there
    return result


def get_wav_bin(wav_contents: bytes):
    output = []

    if has_wav_header(wav_contents):
        wav_contents = wav_contents[44:]

    output.append(struct.pack('4s', b'RIFF'))
    output.append(struct.pack('I', len(wav_contents) + 44))
    output.append(struct.pack('4s', b'WAVE'))
    output.append(struct.pack('4s', b'fmt '))
    output.append(struct.pack('I', 16))
    output.append(struct.pack('H', 1))
    output.append(struct.pack('H', AUDIO_CHANNELS))
    output.append(struct.pack('I', AUDIO_SAMPLE_RATE))
    output.append(struct.pack('I', int(AUDIO_SAMPLE_RATE * _AUDIO_BITS_PER_SAMPLE * AUDIO_CHANNELS / 8)))
    output.append(struct.pack('H', int(_AUDIO_BITS_PER_SAMPLE * AUDIO_CHANNELS / 8)))
    output.append(struct.pack('H', 16))
    output.append(struct.pack('4s', b'data'))
    output.append(struct.pack('I', len(wav_contents)))
    output.append(wav_contents)

    return b''.join(output)


def rtmp_client(exit_flag):
    ffmpeg_async_process = (
        ffmpeg
        .input(RTMP_URL)
        .output('pipe:', format=AUDIO_FORMAT, acodec=AUDIO_CODEC, ar=AUDIO_SAMPLE_RATE, ac=AUDIO_CHANNELS)
        .run_async(pipe_stdout=True, pipe_stderr=False)
    )

    while not exit_flag.is_set():
        # read only last 10 seconds
        bytes_to_read = int(AUDIO_SAMPLE_RATE * AUDIO_CHANNELS * AUDIO_LENGTH_SEC * _AUDIO_BITS_PER_SAMPLE / 8)
        ffmpeg_output = ffmpeg_async_process.stdout.read(bytes_to_read)
        if len(ffmpeg_output) > 0:
            wave_contents = get_wav_bin(ffmpeg_output)

            sec_since_epoch = math.floor(time.time())
            filename = f'{sec_since_epoch}_{uuid.uuid4()}.wav'
            filepath = os.path.join(WRITING_FOLDER, filename)
            with open(filepath, 'wb') as wave_file:
                wave_file.write(wave_contents)
                wave_file.close()
            new_filepath = os.path.join(UNPROCESSED_FOLDER, filename)
            shutil.move(filepath, new_filepath)
            print('record_audio finished: ' + filename)


def http_client(exit_flag):
    print("http_client running")
    sys.stdout.flush()

    while not exit_flag.is_set():
        unprocessed_files = [os.path.join(UNPROCESSED_FOLDER, element) for element in os.listdir(UNPROCESSED_FOLDER)]
        waiting_files = [os.path.join(WAITING_FOLDER, element) for element in os.listdir(WAITING_FOLDER)]

        for filepath in waiting_files:
            if filepath.endswith('.wav'):
                filename = os.path.basename(filepath)
                task_uuid = filename.split('_')[1].split('.')[0]
                params_task_text_status = {
                    'api_key': API_KEY,
                    'task_uuid': task_uuid
                }

                response_raw = requests.post(API_TASK_TEXT_STATUS_URL, params=params_task_text_status)
                response_json = json.dumps(response_raw.json(), ensure_ascii=False)
                response = TaskStatusResponse.from_json(response_json)

                print(task_uuid, response.request_status)
                if response.is_success:
                    if response.request_status == TaskStatusEnum.READY or response.request_status == TaskStatusEnum.NONE:

                        on_file_processed(filepath, response)

                        filename = os.path.basename(filepath)
                        filepath_processed = os.path.join(PROCESSED_FOLDER, filename)
                        shutil.move(filepath, filepath_processed)
                elif response.error_code != 0:
                    print(f'http_client: Get text task status error ({response.error_code}):')
                    print(response.error_desc)

        for filepath in unprocessed_files:
            if filepath.endswith('.wav'):
                filename = os.path.basename(filepath)

                params_task_submit = {
                    'api_key': API_KEY,
                    'features': API_SUBMIT_TASK_FEATURES,
                    'priority': API_TASK_PRIORITY,
                    'language_code': 'lv',
                    'language_codes': ['lv'],
                    'known_and_unknown_users_count': 1,
                    'known_user_ids': [],  # With [1] saying that user_id doesn't exist
                }

                file = open(filepath, 'rb')
                files = {'file': file}

                response_raw = requests.post(API_TASK_SUBMIT_URL, params=params_task_submit, files=files)
                response_json = json.dumps(response_raw.json())
                response = TaskSubmitResponse.from_json(response_json)

                file.close()

                if response.is_success:
                    filename_start = filename.split('_')[0]
                    filename_with_task_uuid = f'{filename_start}_{response.task_uuid}.wav'
                    filepath_waiting = os.path.join(WAITING_FOLDER, filename_with_task_uuid)
                    shutil.move(filepath, filepath_waiting)

                    print('http_client: Task submitted successfully!')
                else:
                    print(f'http_client: Task submit error ({response.error_code}):')
                    print(response.error_desc)

        time.sleep(API_CALL_INTERVAL_SEC)


def on_file_processed(filepath: str, response: TaskStatusResponse):
    print('on_file_processed called!')

    engine = create_engine(f'sqlite:///{DATABASE_FILE}')
    session = Session(engine)
    filename = os.path.basename(filepath)

    for segment in response.results.segments:
        words = segment.text
        for word_raw in words.split(' '):
            word = word_raw.strip()
            if word:
                word = word.lower()
                # capitalise first letter
                word = word[0].upper() + word[1:]
                existing_entry_words = session.query(Words).filter(Words.word == word).first()
                if not existing_entry_words:
                    file_epoch_time = int(filename.split('_')[0])
                    db_created_datetime = datetime.fromtimestamp(file_epoch_time)

                    new_db_word = Words(source_filename=filename, word=word, language='en', created=db_created_datetime)
                    session.add(new_db_word)
                    session.commit()
                existing_entry_words = session.query(Words).filter(Words.word == word).first()
                
                existing_entry_words_stats = session.query(WordsStats).filter(WordsStats.word_id == existing_entry_words.word_id).first()
                if not existing_entry_words_stats:
                    new_db_word_stat = WordsStats(word_id=existing_entry_words.word_id, count=1)
                    session.add(new_db_word_stat)
                else:
                    existing_entry_words_stats.count += 1
                
                session.commit()


def main():
    os.makedirs(UNPROCESSED_FOLDER, exist_ok=True)
    os.makedirs(PROCESSED_FOLDER, exist_ok=True)
    os.makedirs(WAITING_FOLDER, exist_ok=True)
    os.makedirs(WRITING_FOLDER, exist_ok=True)

    exit_flag = multiprocessing.Event()

    rtmp_wav_process = multiprocessing.Process(target=rtmp_client, args=[exit_flag])
    rtmp_wav_process.start()

    http_client_process = multiprocessing.Process(target=http_client, args=[exit_flag])
    http_client_process.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        exit_flag.set()
        print('\n\nmain: Waiting for processes to finish!\n\n')
        rtmp_wav_process.join()
        http_client_process.join()


if __name__ == '__main__':
    main()
